@@ -3,31 +3,34 @@ import sys
 import time
 import pickle
 import signal
+import asyncio
 import unittest
 import threading
+import concurrent
+import dataclasses
 import multiprocessing
 
 from concurrent.futures import CancelledError, TimeoutError
 
 import pebble
 from pebble import ProcessPool, ProcessExpired
+from pebble.pool.base_pool import PoolStatus
 
 
 # set start method
 supported = False
+mp_context = None
 
-if sys.version_info.major > 2 and sys.version_info.minor > 3:
-    methods = multiprocessing.get_all_start_methods()
-    if 'fork' in methods:
-        try:
-            multiprocessing.set_start_method('fork')
 
-            if multiprocessing.get_start_method() == 'fork':
-                supported = True
-        except RuntimeError:  # child process
-            pass
-else:
-    supported = True
+methods = multiprocessing.get_all_start_methods()
+if 'fork' in methods:
+    try:
+        mp_context = multiprocessing.get_context('fork')
+
+        if mp_context.get_start_method() == 'fork':
+            supported = True
+    except RuntimeError:  # child process
+        pass
 
 
 initarg = 0
@@ -43,7 +46,7 @@ def long_initializer():
 
 
 def broken_initializer():
-    raise Exception("BOOM!")
+    raise BaseException("BOOM!")
 
 
 def function(argument, keyword_argument=0):
@@ -56,7 +59,20 @@ def initializer_function():
 
 
 def error_function():
-    raise Exception("BOOM!")
+    raise BaseException("BOOM!")
+
+
+def return_error_function():
+    return BaseException("BOOM!")
+
+
+@dataclasses.dataclass(frozen=True)
+class FrozenError(Exception):
+    pass
+
+
+def frozen_error_function():
+    raise FrozenError()
 
 
 def pickle_error_function():
@@ -82,6 +98,30 @@ def suicide_function():
     os._exit(1)
 
 
+def process_function():
+    p = multiprocessing.Process(target=function, args=[1])
+    p.start()
+    p.join()
+
+    return 1
+
+
+def pool_function():
+    pool = multiprocessing.Pool(1)
+    result = pool.apply(function, args=[1])
+    pool.close()
+    pool.join()
+
+    return result
+
+
+def pebble_function():
+    with ProcessPool(max_workers=1) as pool:
+        f = pool.schedule(function, args=[1])
+
+    return f.result()
+
+
 @unittest.skipIf(not supported, "Start method is not supported")
 class TestProcessPool(unittest.TestCase):
     def setUp(self):
@@ -95,14 +135,14 @@ class TestProcessPool(unittest.TestCase):
     def callback(self, future):
         try:
             self.result = future.result()
-        except Exception as error:
+        except BaseException as error:
             self.exception = error
         finally:
             self.event.set()
 
     def test_process_pool_single_future(self):
         """Process Pool Fork single future."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(function, args=[1],
                                    kwargs={'keyword_argument': 1})
         self.assertEqual(future.result(), 2)
@@ -110,14 +150,14 @@ class TestProcessPool(unittest.TestCase):
     def test_process_pool_multiple_futures(self):
         """Process Pool Fork multiple futures."""
         futures = []
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=2, context=mp_context) as pool:
             for _ in range(5):
                 futures.append(pool.schedule(function, args=[1]))
         self.assertEqual(sum([f.result() for f in futures]), 5)
 
     def test_process_pool_callback(self):
         """Process Pool Fork result is forwarded to the callback."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(
                 function, args=[1], kwargs={'keyword_argument': 1})
         future.add_done_callback(self.callback)
@@ -126,41 +166,53 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_error(self):
         """Process Pool Fork errors are raised by future get."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(error_function)
-        self.assertRaises(Exception, future.result)
+        self.assertRaises(BaseException, future.result)
+
+    def test_process_pool_error_returned(self):
+        """Process Pool Fork returned errors are returned by future get."""
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            future = pool.schedule(return_error_function)
+        self.assertIsInstance(future.result(), BaseException)
 
     def test_process_pool_error_callback(self):
         """Process Pool Fork errors are forwarded to callback."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(error_function)
         future.add_done_callback(self.callback)
         self.event.wait()
-        self.assertTrue(isinstance(self.exception, Exception))
+        self.assertTrue(isinstance(self.exception, BaseException))
 
     def test_process_pool_pickling_error_task(self):
         """Process Pool Fork task pickling errors
         are raised by future.result."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(function, args=[threading.Lock()])
             self.assertRaises((pickle.PicklingError, TypeError), future.result)
 
     def test_process_pool_pickling_error_result(self):
         """Process Pool Fork result pickling errors
         are raised by future.result."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(pickle_error_function)
             self.assertRaises((pickle.PicklingError, TypeError), future.result)
 
+    def test_process_pool_frozen_error(self):
+        """Process Pool Fork frozen errors are raised by future get."""
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            future = pool.schedule(frozen_error_function)
+        self.assertRaises(FrozenError, future.result)
+
     def test_process_pool_timeout(self):
         """Process Pool Fork future raises TimeoutError if so."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(long_function, timeout=0.1)
         self.assertRaises(TimeoutError, future.result)
 
     def test_process_pool_timeout_callback(self):
         """Process Pool Fork TimeoutError is forwarded to callback."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(long_function, timeout=0.1)
         future.add_done_callback(self.callback)
         self.event.wait()
@@ -168,7 +220,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_cancel(self):
         """Process Pool Fork future raises CancelledError if so."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(long_function)
             time.sleep(0.1)  # let the process pick up the task
             self.assertTrue(future.cancel())
@@ -176,7 +228,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_cancel_callback(self):
         """Process Pool Fork CancelledError is forwarded to callback."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(long_function)
             future.add_done_callback(self.callback)
             time.sleep(0.1)  # let the process pick up the task
@@ -184,10 +236,11 @@ class TestProcessPool(unittest.TestCase):
         self.event.wait()
         self.assertTrue(isinstance(self.exception, CancelledError))
 
+    @unittest.skipIf(sys.platform == 'darwin', "Not supported on MAC OS")
     def test_process_pool_different_process(self):
         """Process Pool Fork multiple futures are handled by different processes."""
         futures = []
-        with ProcessPool(max_workers=2) as pool:
+        with ProcessPool(max_workers=2, context=mp_context) as pool:
             for _ in range(0, 5):
                 futures.append(pool.schedule(pid_function))
         self.assertEqual(len(set([f.result() for f in futures])), 2)
@@ -195,14 +248,14 @@ class TestProcessPool(unittest.TestCase):
     def test_process_pool_future_limit(self):
         """Process Pool Fork tasks limit is honored."""
         futures = []
-        with ProcessPool(max_workers=1, max_tasks=2) as pool:
+        with ProcessPool(max_workers=1, max_tasks=2, context=mp_context) as pool:
             for _ in range(0, 4):
                 futures.append(pool.schedule(pid_function))
         self.assertEqual(len(set([f.result() for f in futures])), 2)
 
     def test_process_pool_stop_timeout(self):
         """Process Pool Fork workers are stopped if future timeout."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future1 = pool.schedule(pid_function)
             pool.schedule(long_function, timeout=0.1)
             future2 = pool.schedule(pid_function)
@@ -210,7 +263,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_stop_cancel(self):
         """Process Pool Fork workers are stopped if future is cancelled."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future1 = pool.schedule(pid_function)
             cancel_future = pool.schedule(long_function)
             time.sleep(0.1)  # let the process pick up the task
@@ -220,34 +273,34 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_initializer(self):
         """Process Pool Fork initializer is correctly run."""
-        with ProcessPool(initializer=initializer, initargs=[1]) as pool:
+        with ProcessPool(initializer=initializer, initargs=[1], context=mp_context) as pool:
             future = pool.schedule(initializer_function)
         self.assertEqual(future.result(), 1)
 
     def test_process_pool_broken_initializer(self):
         """Process Pool Fork broken initializer is notified."""
         with self.assertRaises(RuntimeError):
-            with ProcessPool(initializer=broken_initializer) as pool:
+            with ProcessPool(initializer=broken_initializer, context=mp_context) as pool:
                 pool.active
                 time.sleep(0.4)
                 pool.schedule(function)
 
     def test_process_pool_running(self):
         """Process Pool Fork is active if a future is scheduled."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             pool.schedule(function, args=[1])
             self.assertTrue(pool.active)
 
     def test_process_pool_stopped(self):
         """Process Pool Fork is not active once stopped."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             pool.schedule(function, args=[1])
         self.assertFalse(pool.active)
 
     def test_process_pool_close_futures(self):
         """Process Pool Fork all futures are performed on close."""
         futures = []
-        pool = ProcessPool(max_workers=1)
+        pool = ProcessPool(max_workers=1, context=mp_context)
         for index in range(10):
             futures.append(pool.schedule(function, args=[index]))
         pool.close()
@@ -256,7 +309,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_close_stopped(self):
         """Process Pool Fork is stopped after close."""
-        pool = ProcessPool(max_workers=1)
+        pool = ProcessPool(max_workers=1, context=mp_context)
         pool.schedule(function, args=[1])
         pool.close()
         pool.join()
@@ -265,7 +318,7 @@ class TestProcessPool(unittest.TestCase):
     def test_process_pool_stop_futures(self):
         """Process Pool Fork not all futures are performed on stop."""
         futures = []
-        pool = ProcessPool(max_workers=1)
+        pool = ProcessPool(max_workers=1, context=mp_context)
         for index in range(10):
             futures.append(pool.schedule(function, args=[index]))
         pool.stop()
@@ -274,7 +327,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_stop_stopped(self):
         """Process Pool Fork is stopped after stop."""
-        pool = ProcessPool(max_workers=1)
+        pool = ProcessPool(max_workers=1, context=mp_context)
         pool.schedule(function, args=[1])
         pool.stop()
         pool.join()
@@ -282,7 +335,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_stop_stopped_callback(self):
         """Process Pool Fork is stopped in callback."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             def stop_pool_callback(_):
                 pool.stop()
 
@@ -297,9 +350,9 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_large_data(self):
         """Process Pool Fork large data is sent on the channel."""
-        data = "a" * 1098 * 1024 * 50  # 50 Mb
+        data = "a" * 1098 * 1024 * 100  # 100 Mb
 
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(
                 function, args=[data], kwargs={'keyword_argument': ''})
 
@@ -307,9 +360,10 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_stop_large_data(self):
         """Process Pool Fork is stopped if large data is sent on the channel."""
-        data = "a" * 1098 * 1024 * 50  # 50 Mb
-        pool = ProcessPool(max_workers=1)
+        data = "a" * 1098 * 1024 * 100  # 100 Mb
+        pool = ProcessPool(max_workers=1, context=mp_context)
         pool.schedule(function, args=[data])
+        time.sleep(1)
         pool.stop()
         pool.join()
 
@@ -317,7 +371,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_join_workers(self):
         """Process Pool Fork no worker is running after join."""
-        pool = ProcessPool(max_workers=4)
+        pool = ProcessPool(max_workers=4, context=mp_context)
         pool.schedule(function, args=[1])
         pool.stop()
         pool.join()
@@ -325,13 +379,13 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_join_running(self):
         """Process Pool Fork RuntimeError is raised if active pool joined."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             pool.schedule(function, args=[1])
             self.assertRaises(RuntimeError, pool.join)
 
     def test_process_pool_join_futures_timeout(self):
         """Process Pool Fork TimeoutError is raised if join on long futures."""
-        pool = ProcessPool(max_workers=1)
+        pool = ProcessPool(max_workers=1, context=mp_context)
         for _ in range(2):
             pool.schedule(long_function)
         pool.close()
@@ -341,7 +395,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_callback_error(self):
         """Process Pool Fork does not stop if error in callback."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(function, args=[1],
                                    kwargs={'keyword_argument': 1})
             future.add_done_callback(self.callback)
@@ -351,12 +405,12 @@ class TestProcessPool(unittest.TestCase):
                           kwargs={'keyword_argument': 1})
 
     def test_process_pool_exception_isolated(self):
-        """Process Pool Fork an Exception does not affect other futures."""
-        with ProcessPool(max_workers=1) as pool:
+        """Process Pool Fork an BaseException does not affect other futures."""
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(error_function)
             try:
                 future.result()
-            except Exception:
+            except BaseException:
                 pass
             future = pool.schedule(function, args=[1],
                                    kwargs={'keyword_argument': 1})
@@ -365,22 +419,26 @@ class TestProcessPool(unittest.TestCase):
     @unittest.skipIf(os.name == 'nt', "Test won't run on Windows'.")
     def test_process_pool_ignoring_sigterm(self):
         """Process Pool Fork ignored SIGTERM signal are handled on Unix."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(sigterm_function, timeout=0.2)
             with self.assertRaises(TimeoutError):
                 future.result()
 
     def test_process_pool_expired_worker(self):
         """Process Pool Fork unexpect death of worker raises ProcessExpired."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.schedule(suicide_function)
-            self.assertRaises(ProcessExpired, future.result)
+            worker_pid = list(pool._pool_manager.worker_manager.workers)[0]
+            with self.assertRaises(ProcessExpired) as exc_ctx:
+                future.result()
+            self.assertEqual(exc_ctx.exception.exitcode, 1)
+            self.assertEqual(exc_ctx.exception.pid, worker_pid)
 
     def test_process_pool_map(self):
         """Process Pool Fork map simple."""
         elements = [1, 2, 3]
 
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.map(function, elements)
             generator = future.result()
             self.assertEqual(list(generator), elements)
@@ -389,7 +447,7 @@ class TestProcessPool(unittest.TestCase):
         """Process Pool Fork map no elements."""
         elements = []
 
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.map(function, elements)
             generator = future.result()
             self.assertEqual(list(generator), elements)
@@ -398,7 +456,7 @@ class TestProcessPool(unittest.TestCase):
         """Process Pool Fork map one element."""
         elements = [0]
 
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.map(function, elements)
             generator = future.result()
             self.assertEqual(list(generator), elements)
@@ -407,7 +465,7 @@ class TestProcessPool(unittest.TestCase):
         """Process Pool Fork map multiple iterables."""
         expected = (2, 4)
 
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.map(function, (1, 2, 3), (1, 2))
             generator = future.result()
             self.assertEqual(tuple(generator), expected)
@@ -416,14 +474,14 @@ class TestProcessPool(unittest.TestCase):
         """Process Pool Fork map chunksize 1."""
         elements = [1, 2, 3]
 
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.map(function, elements, chunksize=1)
             generator = future.result()
             self.assertEqual(list(generator), elements)
 
     def test_process_pool_map_zero_chunk(self):
         """Process Pool Fork map chunksize 0."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             with self.assertRaises(ValueError):
                 pool.map(function, [], chunksize=0)
 
@@ -432,7 +490,7 @@ class TestProcessPool(unittest.TestCase):
         raised = []
         elements = [1, 2, 3]
 
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.map(long_function, elements, timeout=0.1)
             generator = future.result()
             while True:
@@ -447,12 +505,12 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_map_timeout_chunks(self):
         """Process Pool Fork map timeout is assigned per chunk."""
-        elements = [0.1]*10
+        elements = [0.1]*20
 
-        with ProcessPool(max_workers=1) as pool:
-            # it takes 0.5s to process a chunk
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            # it takes 1s to process a chunk
             future = pool.map(
-                long_function, elements, chunksize=5, timeout=0.8)
+                long_function, elements, chunksize=5, timeout=1.8)
             generator = future.result()
             self.assertEqual(list(generator), elements)
 
@@ -461,7 +519,7 @@ class TestProcessPool(unittest.TestCase):
         raised = None
         elements = [1, 'a', 3]
 
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.map(function, elements)
             generator = future.result()
             while True:
@@ -477,7 +535,7 @@ class TestProcessPool(unittest.TestCase):
 
     def test_process_pool_map_cancel(self):
         """Process Pool Fork cancel iteration."""
-        with ProcessPool(max_workers=1) as pool:
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
             future = pool.map(long_function, range(5))
             generator = future.result()
 
@@ -489,6 +547,235 @@ class TestProcessPool(unittest.TestCase):
                 with self.assertRaises(CancelledError):
                     next(generator)
 
+    def test_process_pool_map_broken_pool(self):
+        """Process Pool Fork Broken Pool."""
+        elements = [1, 2, 3]
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            future = pool.map(long_function, elements, timeout=1)
+            generator = future.result()
+            pool._context.status = PoolStatus.ERROR
+            while True:
+                try:
+                    next(generator)
+                except TimeoutError as error:
+                    self.assertFalse(pool.active)
+                    future.cancel()
+                    break
+                except StopIteration:
+                    break
+
+    def test_process_pool_child_process(self):
+        """Process Pool Fork worker starts process."""
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            future = pool.schedule(process_function)
+        self.assertEqual(future.result(), 1)
+
+    def test_process_pool_child_pool(self):
+        """Process Pool Fork worker starts multiprocessing.Pool."""
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            future = pool.schedule(pool_function)
+        self.assertEqual(future.result(), 1)
+
+    def test_process_pool_child_pebble(self):
+        """Process Pool Fork worker starts pebble.ProcessPool."""
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            future = pool.schedule(pebble_function)
+        self.assertEqual(future.result(), 1)
+
+
+@unittest.skipIf(not supported, "Start method is not supported")
+class TestAsyncIOProcessPool(unittest.TestCase):
+    def setUp(self):
+        self.event = None
+        self.result = None
+        self.exception = None
+
+    def callback(self, future):
+        try:
+            self.result = future.result()
+        # asyncio.exception.CancelledError does not inherit from BaseException
+        except BaseException as error:
+            self.exception = error
+        finally:
+            self.event.set()
+
+    def test_process_pool_single_future(self):
+        """Process Pool Fork single future."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+
+            return await loop.run_in_executor(pool, function, None, 1)
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            self.assertEqual(asyncio.run(test(pool)), 1)
+
+    def test_process_pool_multiple_futures(self):
+        """Process Pool Fork multiple futures."""
+        async def test(pool):
+            futures = []
+            loop = asyncio.get_running_loop()
+
+            for _ in range(5):
+                futures.append(loop.run_in_executor(pool, function, None, 1))
+
+            return await asyncio.wait(futures)
+
+        with ProcessPool(max_workers=2, context=mp_context) as pool:
+            self.assertEqual(sum(r.result()
+                                 for r in asyncio.run(test(pool))[0]), 5)
+
+    def test_process_pool_callback(self):
+        """Process Pool Fork result is forwarded to the callback."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+
+            self.event = asyncio.Event()
+            self.event.clear()
+
+            future = loop.run_in_executor(pool, function, None, 1)
+            future.add_done_callback(self.callback)
+
+            await self.event.wait()
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            asyncio.run(test(pool))
+            self.assertEqual(self.result, 1)
+
+    def test_process_pool_error(self):
+        """Process Pool Fork errors are raised by future get."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+
+            return await loop.run_in_executor(pool, error_function, None)
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            with self.assertRaises(BaseException):
+                asyncio.run(test(pool))
+
+    def test_process_pool_error_returned(self):
+        """Process Pool Fork returned errors are returned by future get."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+
+            return await loop.run_in_executor(pool, return_error_function, None)
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            self.assertIsInstance(asyncio.run(test(pool)), BaseException)
+
+    def test_process_pool_error_callback(self):
+        """Process Pool Fork errors are forwarded to callback."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+
+            self.event = asyncio.Event()
+            self.event.clear()
+
+            future = loop.run_in_executor(pool, error_function, None)
+            future.add_done_callback(self.callback)
+
+            await self.event.wait()
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            asyncio.run(test(pool))
+            self.assertTrue(isinstance(self.exception, BaseException))
+
+    def test_process_pool_timeout(self):
+        """Process Pool Fork future raises TimeoutError if so."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+
+            return await loop.run_in_executor(pool, long_function, 0.1)
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            with self.assertRaises(asyncio.TimeoutError):
+                asyncio.run(test(pool))
+
+    def test_process_pool_timeout_callback(self):
+        """Process Pool Fork TimeoutError is forwarded to callback."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+
+            self.event = asyncio.Event()
+            self.event.clear()
+
+            future = loop.run_in_executor(pool, long_function, 0.1)
+            future.add_done_callback(self.callback)
+
+            await asyncio.sleep(0.1) # let the process pick up the task
+
+            await self.event.wait()
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            asyncio.run(test(pool))
+            self.assertTrue(isinstance(self.exception, asyncio.TimeoutError))
+
+    def test_process_pool_cancel(self):
+        """Process Pool Fork future raises CancelledError if so."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+            future = loop.run_in_executor(pool, long_function, None)
+
+            await asyncio.sleep(0.1) # let the process pick up the task
+
+            self.assertTrue(future.cancel())
+
+            return await future
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            with self.assertRaises(asyncio.CancelledError):
+                asyncio.run(test(pool))
+
+    def test_process_pool_cancel_callback(self):
+        """Process Pool Fork CancelledError is forwarded to callback."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+
+            self.event = asyncio.Event()
+            self.event.clear()
+
+            future = loop.run_in_executor(pool, long_function, None)
+            future.add_done_callback(self.callback)
+
+            await asyncio.sleep(0.1) # let the process pick up the task
+
+            self.assertTrue(future.cancel())
+
+            await self.event.wait()
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            asyncio.run(test(pool))
+            self.assertTrue(isinstance(self.exception, asyncio.CancelledError))
+
+    def test_process_pool_stop_timeout(self):
+        """Process Pool Fork workers are stopped if future timeout."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+            future1 = loop.run_in_executor(pool, pid_function, None)
+            with self.assertRaises(asyncio.TimeoutError):
+                await loop.run_in_executor(pool, long_function, 0.1)
+            future2 = loop.run_in_executor(pool, pid_function, None)
+
+            self.assertNotEqual(await future1, await future2)
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            asyncio.run(test(pool))
+
+    def test_process_pool_stop_cancel(self):
+        """Process Pool Fork workers are stopped if future is cancelled."""
+        async def test(pool):
+            loop = asyncio.get_running_loop()
+            future1 = loop.run_in_executor(pool, pid_function, None)
+            cancel_future = loop.run_in_executor(pool, long_function, None)
+            await asyncio.sleep(0.1)  # let the process pick up the task
+            self.assertTrue(cancel_future.cancel())
+            future2 = loop.run_in_executor(pool, pid_function, None)
+
+            self.assertNotEqual(await future1, await future2)
+
+        with ProcessPool(max_workers=1, context=mp_context) as pool:
+            asyncio.run(test(pool))
+
 
 # DEADLOCK TESTS
 
@@ -514,19 +801,19 @@ class TestProcessPoolDeadlockOnNewFutures(unittest.TestCase):
     def setUp(self):
         self.worker_process = pebble.pool.process.worker_process
         pebble.pool.process.worker_process = broken_worker_process_tasks
-        pebble.pool.channel.LOCK_TIMEOUT = 0.1
+        pebble.CONSTS.channel_lock_timeout = 0.1
 
     def tearDown(self):
         pebble.pool.process.worker_process = self.worker_process
-        pebble.pool.channel.LOCK_TIMEOUT = 60
+        pebble.CONSTS.channel_lock_timeout = 60
 
     def test_pool_deadlock_stop(self):
         """Process Pool Fork reading deadlocks are stopping the Pool."""
         with self.assertRaises(RuntimeError):
-            pool = pebble.ProcessPool(max_workers=1)
+            pool = pebble.ProcessPool(max_workers=1, context=mp_context)
             for _ in range(10):
                 pool.schedule(function)
-                time.sleep(0.1)
+                time.sleep(0.2)
 
 
 @unittest.skipIf(not supported, "Start method is not supported")
@@ -534,22 +821,31 @@ class TestProcessPoolDeadlockOnResult(unittest.TestCase):
     def setUp(self):
         self.worker_process = pebble.pool.process.worker_process
         pebble.pool.process.worker_process = broken_worker_process_result
-        pebble.pool.channel.LOCK_TIMEOUT = 0.1
+        pebble.CONSTS.channel_lock_timeout = 0.1
 
     def tearDown(self):
         pebble.pool.process.worker_process = self.worker_process
-        pebble.pool.channel.LOCK_TIMEOUT = 60
+        pebble.CONSTS.channel_lock_timeout = 60
 
     def test_pool_deadlock(self):
         """Process Pool Fork no deadlock if writing worker dies locking channel."""
-        with pebble.ProcessPool(max_workers=1) as pool:
+        with pebble.ProcessPool(max_workers=1, context=mp_context) as pool:
             with self.assertRaises(pebble.ProcessExpired):
                 pool.schedule(function).result()
 
-    def test_pool_deadlock_stop(self):
-        """Process Pool Fork writing deadlocks are stopping the Pool."""
-        with self.assertRaises(RuntimeError):
-            pool = pebble.ProcessPool(max_workers=1)
-            for _ in range(10):
-                pool.schedule(function)
-                time.sleep(0.1)
+
+@unittest.skipIf(not supported, "Start method is not supported")
+class TestProcessPoolDeadlockOnCancelLargeData(unittest.TestCase):
+    def test_pool_deadlock_stop_cancel(self):
+        """Pool is stopped when futures are cancelled on large data."""
+        data = b'A' * 1024 * 1024 * 100
+
+        with pebble.ProcessPool() as pool:
+            futures = [pool.schedule(function, args=[data]) for _ in range(10)]
+            concurrent.futures.wait(
+                futures,
+                return_when=concurrent.futures.FIRST_COMPLETED
+            )
+            for f in futures:
+                f.cancel()
+            pool.stop()
